@@ -1,17 +1,16 @@
-;;; magit-merge.el --- merge functionality  -*- lexical-binding: t -*-
+;;; magit-merge.el --- Merge functionality  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2010-2018  The Magit Project Contributors
-;;
-;; You should have received a copy of the AUTHORS.md file which
-;; lists all contributors.  If not, see http://magit.vc/authors.
+;; Copyright (C) 2008-2025 The Magit Project Contributors
 
-;; Author: Jonas Bernoulli <jonas@bernoul.li>
-;; Maintainer: Jonas Bernoulli <jonas@bernoul.li>
+;; Author: Jonas Bernoulli <emacs.magit@jonas.bernoulli.dev>
+;; Maintainer: Jonas Bernoulli <emacs.magit@jonas.bernoulli.dev>
 
-;; Magit is free software; you can redistribute it and/or modify it
+;; SPDX-License-Identifier: GPL-3.0-or-later
+
+;; Magit is free software: you can redistribute it and/or modify it
 ;; under the terms of the GNU General Public License as published by
-;; the Free Software Foundation; either version 3, or (at your option)
-;; any later version.
+;; the Free Software Foundation, either version 3 of the License, or
+;; (at your option) any later version.
 ;;
 ;; Magit is distributed in the hope that it will be useful, but WITHOUT
 ;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
@@ -19,7 +18,7 @@
 ;; License for more details.
 ;;
 ;; You should have received a copy of the GNU General Public License
-;; along with Magit.  If not, see http://www.gnu.org/licenses.
+;; along with Magit.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
@@ -27,34 +26,63 @@
 
 ;;; Code:
 
-(eval-when-compile
-  (require 'subr-x))
-
 (require 'magit)
+(require 'magit-diff)
 
 (declare-function magit-git-push "magit-push" (branch target args))
 
 ;;; Commands
 
-;;;###autoload (autoload 'magit-merge-popup "magit" nil t)
-(magit-define-popup magit-merge-popup
-  "Popup console for merge commands."
+;;;###autoload (autoload 'magit-merge "magit" nil t)
+(transient-define-prefix magit-merge ()
+  "Merge branches."
   :man-page "git-merge"
-  :switches '((?f "Fast-forward only" "--ff-only")
-              (?n "No fast-forward"   "--no-ff"))
-  :options  '((?s "Strategy" "--strategy="))
-  :actions  '((?m "Merge"                  magit-merge-plain)
-              (?p "Preview merge"          magit-merge-preview)
-              (?e "Merge and edit message" magit-merge-editmsg) nil
-              (?n "Merge but don't commit" magit-merge-nocommit)
-              (?s "Squash merge"           magit-merge-squash)
-              (?a "Absorb"                 magit-merge-absorb)
-              (?i "Merge into"             magit-merge-into))
-  :sequence-actions   '((?m "Commit merge" magit-commit-create)
-                        (?a "Abort merge"  magit-merge-abort))
-  :sequence-predicate 'magit-merge-in-progress-p
-  :default-action 'magit-merge
-  :max-action-columns 2)
+  :incompatible '(("--ff-only" "--no-ff"))
+  ["Arguments"
+   :if-not magit-merge-in-progress-p
+   ("-f" "Fast-forward only" "--ff-only")
+   ("-n" "No fast-forward"   "--no-ff")
+   (magit-merge:--strategy)
+   (5 magit-merge:--strategy-option)
+   (5 "-b" "Ignore changes in amount of whitespace" "-Xignore-space-change")
+   (5 "-w" "Ignore whitespace when comparing lines" "-Xignore-all-space")
+   (5 magit-diff:--diff-algorithm :argument "-Xdiff-algorithm=")
+   (5 magit:--gpg-sign)]
+  ["Actions"
+   :if-not magit-merge-in-progress-p
+   [("m" "Merge"                  magit-merge-plain)
+    ("e" "Merge and edit message" magit-merge-editmsg)
+    ("n" "Merge but don't commit" magit-merge-nocommit)
+    ("a" "Absorb"                 magit-merge-absorb)]
+   [("p" "Preview merge"          magit-merge-preview)
+    ""
+    ("s" "Squash merge"           magit-merge-squash)
+    ("i" "Dissolve"               magit-merge-into)]]
+  ["Actions"
+   :if magit-merge-in-progress-p
+   ("m" "Commit merge" magit-commit-create)
+   ("a" "Abort merge"  magit-merge-abort)])
+
+(defun magit-merge-arguments ()
+  (transient-args 'magit-merge))
+
+(transient-define-argument magit-merge:--strategy ()
+  :description "Strategy"
+  :class 'transient-option
+  ;; key for merge and rebase: "-s"
+  ;; key for cherry-pick and revert: "=s"
+  ;; shortarg for merge and rebase: "-s"
+  ;; shortarg for cherry-pick and revert: none
+  :key "-s"
+  :argument "--strategy="
+  :choices '("resolve" "recursive" "octopus" "ours" "subtree"))
+
+(transient-define-argument magit-merge:--strategy-option ()
+  :description "Strategy Option"
+  :class 'transient-option
+  :key "-X"
+  :argument "--strategy-option="
+  :choices '("ours" "theirs" "patience"))
 
 ;;;###autoload
 (defun magit-merge-plain (rev &optional args nocommit)
@@ -84,7 +112,8 @@ edit it.
   (magit-merge-assert)
   (cl-pushnew "--no-ff" args :test #'equal)
   (apply #'magit-run-git-with-editor "merge" "--edit"
-         (append args (list rev))))
+         (append (delete "--ff-only" args)
+                 (list rev))))
 
 ;;;###autoload
 (defun magit-merge-nocommit (rev &optional args)
@@ -106,19 +135,24 @@ Before merging, force push the source branch to its push-remote,
 provided the respective remote branch already exists, ensuring
 that the respective pull-request (if any) won't get stuck on some
 obsolete version of the commits that are being merged.  Finally
-if `magit-branch-pull-request' was used to create the merged
-branch, then also remove the respective remote branch."
+if `forge-branch-pullreq' was used to create the merged branch,
+then also remove the respective remote branch."
   (interactive
    (list (magit-read-other-local-branch
-          (format "Merge `%s' into" (magit-get-current-branch))
+          (format "Merge `%s' into"
+                  (or (magit-get-current-branch)
+                      (magit-rev-parse "HEAD")))
           nil
-          (when-let ((upstream (magit-get-upstream-branch)))
-            (when-let ((upstream (cdr (magit-split-branch-name upstream))))
-              (and (magit-branch-p upstream) upstream))))
+          (and-let* ((upstream (magit-get-upstream-branch))
+                     (upstream (cdr (magit-split-branch-name upstream))))
+            (and (magit-branch-p upstream) upstream)))
          (magit-merge-arguments)))
-  (let ((current (magit-get-current-branch)))
+  (let ((current (magit-get-current-branch))
+        (head (magit-rev-parse "HEAD")))
     (when (zerop (magit-call-git "checkout" branch))
-      (magit--merge-absort current args))))
+      (if current
+          (magit--merge-absorb current args)
+        (magit-run-git-with-editor "merge" args head)))))
 
 ;;;###autoload
 (defun magit-merge-absorb (branch &optional args)
@@ -128,16 +162,17 @@ Before merging, force push the source branch to its push-remote,
 provided the respective remote branch already exists, ensuring
 that the respective pull-request (if any) won't get stuck on some
 obsolete version of the commits that are being merged.  Finally
-if `magit-branch-pull-request' was used to create the merged
-branch, then also remove the respective remote branch."
+if `forge-branch-pullreq' was used to create the merged branch,
+then also remove the respective remote branch."
   (interactive (list (magit-read-other-local-branch "Absorb branch")
                      (magit-merge-arguments)))
-  (magit--merge-absort branch args))
+  (magit--merge-absorb branch args))
 
-(defun magit--merge-absort (branch args)
-  (when (equal branch "master")
+(defun magit--merge-absorb (branch args)
+  (when (equal branch (magit-main-branch))
     (unless (yes-or-no-p
-             "Do you really want to merge `master' into another branch? ")
+             (format "Do you really want to merge `%s' into another branch? "
+                     branch))
       (user-error "Abort")))
   (if-let ((target (magit-get-push-branch branch t)))
       (progn
@@ -150,11 +185,22 @@ branch, then also remove the respective remote branch."
                  (magit-process-sentinel process event)
                (process-put process 'inhibit-refresh t)
                (magit-process-sentinel process event)
-               (magit--merge-absort-1 branch args))))))
-    (magit--merge-absort-1 branch args)))
+               (magit--merge-absorb-1 branch args))))))
+    (magit--merge-absorb-1 branch args)))
 
-(defun magit--merge-absort-1 (branch args)
-  (magit-run-git-async "merge" args "--no-edit" branch)
+(defun magit--merge-absorb-1 (branch args)
+  (if-let ((pr (magit-get "branch" branch "pullRequest")))
+      (magit-run-git-async
+       "merge" args "-m"
+       (format "Merge branch '%s'%s [#%s]"
+               branch
+               (let ((current (magit-get-current-branch)))
+                 (if (equal current (magit-main-branch))
+                     ""
+                   (format " into %s" current)))
+               pr)
+       branch)
+    (magit-run-git-async "merge" args "--no-edit" branch))
   (set-process-sentinel
    magit-this-process
    (lambda (process event)
@@ -179,29 +225,14 @@ branch, then also remove the respective remote branch."
 (defun magit-merge-preview (rev)
   "Preview result of merging REV into the current branch."
   (interactive (list (magit-read-other-branch-or-commit "Preview merge")))
-  (magit-mode-setup #'magit-merge-preview-mode rev))
-
-(define-derived-mode magit-merge-preview-mode magit-diff-mode "Magit Merge"
-  "Mode for previewing a merge."
-  :group 'magit-diff
-  (hack-dir-local-variables-non-file-buffer))
-
-(defun magit-merge-preview-refresh-buffer (rev)
-  (let* ((branch (magit-get-current-branch))
-         (head (or branch (magit-rev-verify "HEAD"))))
-    (magit-set-header-line-format (format "Preview merge of %s into %s"
-                                          rev
-                                          (or branch "HEAD")))
-    (magit-insert-section (diffbuf)
-      (magit-git-wash #'magit-diff-wash-diffs
-        "merge-tree" (magit-git-string "merge-base" head rev) head rev))))
+  (magit-merge-preview-setup-buffer rev))
 
 ;;;###autoload
 (defun magit-merge-abort ()
   "Abort the current merge operation.
 \n(git merge --abort)"
   (interactive)
-  (unless (file-exists-p (magit-git-dir "MERGE_HEAD"))
+  (unless (file-exists-p (expand-file-name "MERGE_HEAD" (magit-gitdir)))
     (user-error "No merge in progress"))
   (magit-confirm 'abort-merge)
   (magit-run-git-async "merge" "--abort"))
@@ -221,7 +252,9 @@ branch, then also remove the respective remote branch."
             (user-error "Quit")))))
   (pcase (cons arg (cddr (car (magit-file-status file))))
     ((or `("--ours"   ?D ,_)
-         `("--theirs" ,_ ?D))
+         '("--ours"   ?U ?A)
+         `("--theirs" ,_ ?D)
+         '("--theirs" ?A ?U))
      (magit-run-git "rm" "--" file))
     (_ (if (equal arg "--merge")
            ;; This fails if the file was deleted on one
@@ -233,12 +266,13 @@ branch, then also remove the respective remote branch."
 ;;; Utilities
 
 (defun magit-merge-in-progress-p ()
-  (file-exists-p (magit-git-dir "MERGE_HEAD")))
+  (file-exists-p (expand-file-name "MERGE_HEAD" (magit-gitdir))))
 
 (defun magit--merge-range (&optional head)
   (unless head
     (setq head (magit-get-shortname
-                (car (magit-file-lines (magit-git-dir "MERGE_HEAD"))))))
+                (car (magit-file-lines
+                      (expand-file-name "MERGE_HEAD" (magit-gitdir)))))))
   (and head
        (concat (magit-git-string "merge-base" "--octopus" "HEAD" head)
                ".." head)))
@@ -252,33 +286,29 @@ branch, then also remove the respective remote branch."
   (magit-read-char-case (format "For %s checkout: " file) t
     (?o "[o]ur stage"   "--ours")
     (?t "[t]heir stage" "--theirs")
-    (?c "[c]onflict"    "--merge")))
+    (?c (if magit-verbose-messages "restore [c]onflict" "[c]onflict")
+        "--merge")))
 
 ;;; Sections
 
-(defvar magit-unmerged-section-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map [remap magit-visit-thing] 'magit-diff-dwim)
-    map)
-  "Keymap for `unmerged' sections.")
-
 (defun magit-insert-merge-log ()
   "Insert section for the on-going merge.
 Display the heads that are being merged.
 If no merge is in progress, do nothing."
   (when (magit-merge-in-progress-p)
     (let* ((heads (mapcar #'magit-get-shortname
-                          (magit-file-lines (magit-git-dir "MERGE_HEAD"))))
+                          (magit-file-lines
+                           (expand-file-name "MERGE_HEAD" (magit-gitdir)))))
            (range (magit--merge-range (car heads))))
       (magit-insert-section (unmerged range)
         (magit-insert-heading
-          (format "Merging %s:" (mapconcat #'identity heads ", ")))
-        (magit-insert-log
-         range
-         (let ((args magit-log-section-arguments))
-           (unless (member "--decorate=full" magit-log-section-arguments)
-             (push "--decorate=full" args))
-           args))))))
+          (format "Merging %s:" (string-join heads ", ")))
+        (magit--insert-log nil
+          range
+          (let ((args magit-buffer-log-args))
+            (unless (member "--decorate=full" magit-buffer-log-args)
+              (push "--decorate=full" args))
+            args))))))
 
 ;;; _
 (provide 'magit-merge)
